/*
  Copyright (c) 2012, Matthias Schiffer <mschiffer@universe-factory.net>
  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation
       and/or other materials provided with the distribution.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#include "ffd.h"
#include "netif.h"
#include "packet.h"
#include "tlv.h"
#include "tlv_types.h"

#include <errno.h>
#include <poll.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <netpacket/packet.h>

#include <sys/socket.h>
#include <sys/uio.h>


#define FFD_PROTO 0xffd

static const eth_addr_t ffd_addr = {{0x03, 0x00, 0x00, 0x00, 0x0f, 0xfd}};

#define HELLO_INTERVAL 1000
#define IHU_INTERVAL (3*HELLO_INTERVAL)
#define PACKET_MAX 1000


static char *mesh = "bat0";

static int sockfd;
static struct timespec now;

static ffd_iface_t *iface_list = NULL;

static ffd_neigh_t self;
//static ffd_orig_t own_data;

/* neighs and origs that have been changed must be moved to front */
//static ffd_neigh_t *neigh_data = NULL;
//static ffd_orig_t *orig_data = NULL;


static inline bool use_netif(const char *ifname) {
	char *if_mesh = netif_get_mesh(ifname);
	bool ret = (if_mesh && !strcmp(if_mesh, mesh));
	free(if_mesh);

	return ret;
}

static void update_time() {
	clock_gettime(CLOCK_MONOTONIC, &now);
}

static bool check_config() {
	if (!netif_is_mesh(mesh)) {
		fprintf(stderr, "error: configured interface is no mesh\n");
		return false;
	}

	return true;
}

static bool init_self() {
	eth_addr_t primary_addr = netif_mesh_get_primary_addr(mesh);
	if (is_eth_addr_unspec(&primary_addr))
		return false;

	memset(&self, 0, sizeof(self));
	//memset(&own_data, 0, sizeof(own_data));

	//self.addr = own_data.addr = primary_addr;

	//random_bytes(&self.rev, sizeof(self.rev));
	//random_bytes(&own_data.rev, sizeof(own_data.rev));

	return true;
}

static bool init_socket() {
	sockfd = socket(AF_PACKET, SOCK_DGRAM, htons(FFD_PROTO));
	if (sockfd < 0) {
		fprintf(stderr, "error: socket: %m\n");
		return false;
	}

	return true;
}

static inline ffd_iface_t* get_iface(unsigned ifindex) {
	ffd_iface_t *iface;
	for (iface = iface_list; iface; iface = iface->next) {
		if (iface->ifindex == ifindex)
			return iface;
	}

	return NULL;
}

static void update_netif(const char *ifname, unsigned ifindex, void *arg) {
	if (!use_netif(ifname))
		return;

	/* join multicast group */
	struct packet_mreq mr;
	memset(&mr, 0, sizeof(mr));
	mr.mr_ifindex = ifindex;
	mr.mr_type = PACKET_MR_MULTICAST;
	mr.mr_alen = ETH_ALEN;
	memcpy(mr.mr_address, ffd_addr.d, ETH_ALEN);
	if (setsockopt(sockfd, SOL_PACKET, PACKET_ADD_MEMBERSHIP, &mr, sizeof(mr)) && errno != EADDRINUSE)
		fprintf(stderr, "warning: setsockopt: %m\n");

	ffd_iface_t *iface = get_iface(ifindex);

	if (!iface) {
		/* new iface */
		iface = malloc(sizeof(ffd_iface_t));
		iface->next = iface_list;
		iface_list = iface;

		iface->seqno = 0;
		iface->neigh_list = NULL;
	}

	iface->ifindex = ifindex;
	strncpy(iface->name, ifname, IF_NAMESIZE);
	iface->type = netif_get_type(ifname);
	iface->addr = netif_get_eth_addr(ifname);
}

static inline ffd_neigh_t* get_neigh(const ffd_iface_t *iface, const eth_addr_t *addr) {
	ffd_neigh_t *neigh;
	for (neigh = iface->neigh_list; neigh; neigh = neigh->next) {
		if (are_eth_addrs_equal(addr, &neigh->addr))
			return neigh;
	}

	return neigh;
}

static ffd_neigh_t* get_neigh_create(ffd_iface_t *iface, const eth_addr_t *addr) {
	ffd_neigh_t *neigh = get_neigh(iface, addr);
	if (!neigh) {
		neigh = calloc(1, sizeof(ffd_neigh_t));
		neigh->next = iface->neigh_list;
		iface->neigh_list = neigh;
		neigh->addr = *addr;

		neigh->txcost = 0xffff;
	}

	return neigh;
}

static void free_neighs(ffd_neigh_t *neigh) {
	ffd_neigh_t *next;
	for (; neigh; neigh = next) {
		next = neigh->next;
		free(neigh);
	}
}

static uint16_t get_neigh_rxcost(const ffd_neigh_t *neigh) {
	int timediff = timespec_diff(&now, &neigh->last_hello)/10;
	int shift = (timediff - neigh->hello_interval/2)/neigh->hello_interval;
	int received = __builtin_popcount(neigh->hello_log << shift);

	if (received == 0)
		return 0xffff;
	else
		return (0x1000/received);
}

static void update_netifs() {
	ffd_iface_t *iface, **cur;

	for (iface = iface_list; iface; iface = iface->next)
		iface->type = IF_UNSPEC;

	netif_foreach(update_netif, NULL);

	cur = &iface_list;
	while (*cur) {
		iface = *cur;

		if (iface->type == IF_UNSPEC) {
			*cur = iface->next;

			free_neighs(iface->neigh_list);
			free(iface);
		}
		else {
			cur = &iface->next;
		}
	}
}

static bool send_eth(const eth_addr_t *addr, unsigned ifindex, void *buf, size_t len) {
	static const uint8_t zeros[46] = {0};

	struct sockaddr_ll sa;
	memset(&sa, 0, sizeof(sa));
	sa.sll_family = AF_PACKET;
	sa.sll_protocol = htons(FFD_PROTO);
	sa.sll_ifindex = ifindex;
	sa.sll_halen = ETH_ALEN;
	memcpy(sa.sll_addr, addr->d, ETH_ALEN);

	struct iovec vec[2] = {
		{ .iov_base = buf, .iov_len = len },
		{ .iov_base = (void*)zeros, .iov_len = sizeof(zeros) - len },
	};

	struct msghdr msg;
	memset(&msg, 0, sizeof(msg));
	msg.msg_name = &sa;
	msg.msg_namelen = sizeof(sa);
	msg.msg_iov = vec;
	msg.msg_iovlen = (len < 46) ? 2 : 1;


	while (sendmsg(sockfd, &msg, 0) < 0) {
		if (errno != EINTR)
			return false;
	}

	return true;
}

static void add_ihus(ffd_packet_t *packet, size_t max_len, const ffd_iface_t *iface) {
	const ffd_neigh_t *neigh;

	for (neigh = iface->neigh_list; neigh; neigh = neigh->next) {
		ffd_tlv_ihu_t *ihu = ffd_tlv_add(packet, PACKET_MAX, TLV_IHU, sizeof(ffd_tlv_ihu_t)+sizeof(eth_addr_t));
		if (!ihu)
			return;

		ihu->ae = ADDR_ENC_ETH;
		ihu->reserved = 0;
		ihu->rxcost = htons(get_neigh_rxcost(neigh));
		ihu->interval = htons(IHU_INTERVAL);
		memcpy(ihu->address, &neigh->addr, sizeof(eth_addr_t));
	}
}

static void send_hellos() {
	ffd_packet_t *packet = alloca(sizeof(ffd_packet_t)+PACKET_MAX);

	packet->version_magic = htons(FFD_VERSION_MAGIC);
	packet->len = 0;

	ffd_tlv_hello_t *hello = ffd_tlv_add(packet, PACKET_MAX, TLV_HELLO, sizeof(ffd_tlv_hello_t));
	if (!hello)
		return;

	hello->reserved = 0;
	hello->interval = htons(HELLO_INTERVAL);

	uint16_t len = packet->len;

	ffd_iface_t *iface;
	for (iface = iface_list; iface; iface = iface->next) {
		hello->seqno = htons(iface->seqno++);

		packet->len = len;

		add_ihus(packet, PACKET_MAX, iface);

		if (!send_eth(&ffd_addr, iface->ifindex, packet, sizeof(ffd_packet_t)+ntohs(packet->len)))
			fprintf(stderr, "send_eth: %m\n");
	}
}

static void handle_tlv_hello(const ffd_tlv_hello_t *tlv_hello, size_t len, const eth_addr_t *addr, ffd_iface_t *iface) {
	if (len < sizeof(ffd_tlv_hello_t)) {
		fprintf(stderr, "warn: received short hello TLV.\n");
		return;
	}

	fprintf(stderr, "debug: received hello with seqno %u.\n", ntohs(tlv_hello->seqno));

	ffd_neigh_t *neigh = get_neigh_create(iface, addr);

	uint16_t seqno = ntohs(tlv_hello->seqno);

	if (neigh->last_hello.tv_sec) {
		int timediff = timespec_diff(&now, &neigh->last_hello)/10;
		uint16_t seqexp = neigh->last_seqno + (timediff - neigh->hello_interval/2)/neigh->hello_interval;

		/* cast to int16_t to ensure correct handling of seqno wrapping */
		if (abs((int16_t)(seqno - seqexp)) > 16) {
			fprintf(stderr, "info: neighbour was reset.\n");
			neigh->hello_log = 0;
		}
		else {
			int16_t seqdiff = seqno - neigh->last_seqno;
			if (seqdiff <= 0) {
				fprintf(stderr, "debug: seqno already seen, ignoring.\n");
				return;
			}

			neigh->hello_log <<= seqdiff;
		}
	}
	else {
		fprintf(stderr, "info: received hello from new neighbour.\n");
	}

	neigh->hello_log |= 1;
	neigh->hello_interval = ntohs(tlv_hello->interval);
	neigh->last_seqno = seqno;
	neigh->last_hello = now;

	fprintf(stderr, "debug: accepted hello, log %04x, rxcost is %u now.\n", neigh->hello_log, get_neigh_rxcost(neigh));
}

static void handle_tlv_ihu(const ffd_tlv_ihu_t *tlv_ihu, size_t len, const eth_addr_t *addr, ffd_iface_t *iface) {
	if (len < sizeof(ffd_tlv_ihu_t)) {
		fprintf(stderr, "warn: received short IHU TLV.\n");
		return;
	}

	if (tlv_ihu->ae == ADDR_ENC_ETH) {
		if (len < sizeof(ffd_tlv_ihu_t)+sizeof(eth_addr_t)) {
			fprintf(stderr, "warn: received short IHU TLV.\n");
			return;
		}

		if (memcmp(tlv_ihu->address, &iface->addr, sizeof(eth_addr_t)) != 0)
			return;
	}
	else if (tlv_ihu->ae != ADDR_ENC_UNSPEC) {
		return;
	}

	ffd_neigh_t *neigh = get_neigh_create(iface, addr);
	neigh->last_ihu = now;
	neigh->txcost = ntohs(tlv_ihu->rxcost);

	fprintf(stderr, "debug: accepted IHU, txcost is %u now.\n", neigh->txcost);
}

static void handle_tlv(ffd_tlv_type_t type, const void *data, size_t len, void *arg) {
	const struct sockaddr_ll *from = arg;
	const eth_addr_t *addr = (const eth_addr_t*)from->sll_addr;
	ffd_iface_t *iface = get_iface(from->sll_ifindex);

	if (!iface)
		return;

	switch (type) {
	case TLV_HELLO:
		handle_tlv_hello(data, len, addr, iface);
		return;

	case TLV_IHU:
		handle_tlv_ihu(data, len, addr, iface);
		return;

	default:
		fprintf(stderr, "debug: received unknown TLV %u on %s.\n", type, iface->name);
		return;
	}
}

static void receive_packet() {
	ffd_packet_t *packet = alloca(sizeof(ffd_packet_t)+PACKET_MAX);
	struct sockaddr_ll from;
	socklen_t fromlen = sizeof(from);

	ssize_t readlen = recvfrom(sockfd, packet, sizeof(ffd_packet_t)+PACKET_MAX, 0, (struct sockaddr*)&from, &fromlen);

	if (readlen < 0) {
		fprintf(stderr, "error: recvfrom: %m\n");
		return;
	}

	if (from.sll_halen != ETH_ALEN) {
		fprintf(stderr, "error: receive_packet: strange address length\n");
		return;
	}

	if (readlen < sizeof(ffd_packet_t) || readlen < sizeof(ffd_packet_t)+ntohs(packet->len)) {
		fprintf(stderr, "debug: received short packet.\n");
		return;
	}

	if (!ffd_tlv_parse(packet, handle_tlv, &from)) {
		fprintf(stderr, "debug: received invalid packet.\n");
	}
}

void add_interval(struct timespec *time, int interval) {
	int cs = time->tv_nsec/1e7 + interval;

	time->tv_sec += cs/100;
	time->tv_nsec = time->tv_nsec%(long)1e7 + (cs%100)*1e7;
}

int main() {
	if (!check_config())
		return 1;

	if (!init_self())
		return 1;

	if (!init_socket())
		return 1;

	update_time();

	struct timespec next_hello = now;

	while (true) {
		update_netifs();

		int timeout = timespec_diff(&next_hello, &now);

		if (timeout <= 0) {
			send_hellos();

			add_interval(&next_hello, HELLO_INTERVAL);
			continue;
		}

		struct pollfd fds[1];

		fds[0].fd = sockfd;
		fds[0].events = POLLIN;

		poll(fds, 1, timeout);

		update_time();

		if (fds[0].revents & POLLIN)
			receive_packet();
	}

	return 0;
}

